''******************************
''*  Altair CPU Simulator v1.3 *
''*  (C) 2008 Michael Rychlik  *
''******************************

' v1.0 - 01 Oct 2007 - original version
' v1.1 - 09 Dec 2007 - Moved dispatch table into HUB RAM to save COG space (and increase speed !!)
' v1.2 - 09 Mar 2008 - No change, just bump version 
' v1.3 - 20 Mar 2008 - Change do_xchg and do_xthl to use the xor swap trick with no temp storage.
'                    - memory read write overhaul , checks for RAM/ROM address limits as we don't
'                      want broken 8085 programs being able to trample over HUB RAM
'                    - Removed those flag bits from the dispatch table. Now immediate bytes words
'                      are read by in the opcode handlers that need them not in the main loop.
'                    - Removed defaulting mem_ptr to hl in main instruction loop, instructions that need it
'                      handle it themselves.
'                    - Changed push/pop_word to use memory_read/write functions.
'

CON                                
  
  'Register defs
  reg_b = 0
  reg_c = 1
  reg_d = 2
  reg_e = 3
  reg_h = 4
  reg_l = 5
  reg_m = 6
  reg_a = 7

  'Operation defs for accumulator instructions
  op_add = 0
  op_adc = 1
  op_sub = 2
  op_sbb = 3
  op_ana = 4
  op_xra = 5
  op_ora = 6
  op_cmp = 7
              
  'Register pair defs for register pair instructions
  rp_bc  = 0
  rp_de  = 1
  rp_hl  = 2
  rp_psw = 3
  rp_sp  = rp_psw
  rp_mask= %00110000


  '8080 instruction op codes

  'Carry bit instructions
  cmc   = %00111111
  stc   = %00110111
  
  'Single register instructions
  inr_mask = %11_000_111
  inr      = %00_000_100
  inr_b    = %00_000_100
  inr_c    = %00_001_100
  inr_d    = %00_010_100
  inr_e    = %00_011_100
  inr_h    = %00_100_100
  inr_l    = %00_101_100
  inr_m    = %00_110_100
  inr_a    = %00_111_100

  dcr_mask = %11_000_111
  dcr      = %00_000_101
  dcr_b    = %00_000_101
  dcr_c    = %00_001_101
  dcr_d    = %00_010_101
  dcr_e    = %00_011_101
  dcr_h    = %00_100_101
  dcr_l    = %00_101_101
  dcr_m    = %00_110_101
  dcr_a    = %00_111_101

  'data transfer instructions
  mov_mask     = %11000000
  mov_         = %01000000
  mov_src_mask = %00000111
  mov_dst_mask = %00111000

  mov_b_b      = %01_000_000
  mov_b_c      = %01_000_001
  mov_b_d      = %01_000_010
  mov_b_e      = %01_000_011
  mov_b_h      = %01_000_100
  mov_b_l      = %01_000_101
  mov_b_m      = %01_000_110
  mov_b_a      = %01_000_111

  mov_c_b      = %01_001_000
  mov_c_c      = %01_001_001
  mov_c_d      = %01_001_010
  mov_c_e      = %01_001_011
  mov_c_h      = %01_001_100
  mov_c_l      = %01_001_101
  mov_c_m      = %01_001_110
  mov_c_a      = %01_001_111

  mov_d_b      = %01_010_000
  mov_d_c      = %01_010_001
  mov_d_d      = %01_010_010    
  mov_d_e      = %01_010_011
  mov_d_h      = %01_010_100
  mov_d_l      = %01_010_101
  mov_d_m      = %01_010_110
  mov_d_a      = %01_010_111

  mov_e_b      = %01_011_000
  mov_e_c      = %01_011_001
  mov_e_d      = %01_011_010
  mov_e_e      = %01_011_011
  mov_e_h      = %01_011_100
  mov_e_l      = %01_011_101
  mov_e_m      = %01_011_110
  mov_e_a      = %01_011_111

  mov_h_b      = %01_100_000
  mov_h_c      = %01_100_001
  mov_h_d      = %01_100_010
  mov_h_e      = %01_100_011
  mov_h_h      = %01_100_100
  mov_h_l      = %01_100_101
  mov_h_m      = %01_100_110
  mov_h_a      = %01_100_111

  mov_l_b      = %01_101_000
  mov_l_c      = %01_101_001
  mov_l_d      = %01_101_010
  mov_l_e      = %01_101_011
  mov_l_h      = %01_101_100
  mov_l_l      = %01_101_101
  mov_l_m      = %01_101_110
  mov_l_a      = %01_101_111

  mov_m_b      = %01_110_000
  mov_m_c      = %01_110_001
  mov_m_d      = %01_110_010
  mov_m_e      = %01_110_011
  mov_m_h      = %01_110_100
  mov_m_l      = %01_110_101
  mov_m_m      = %01_110_110
  mov_m_a      = %01_110_111

  mov_a_b      = %01_111_000
  mov_a_c      = %01_111_001
  mov_a_d      = %01_111_010
  mov_a_e      = %01_111_011
  mov_a_h      = %01_111_100
  mov_a_l      = %01_111_101
  mov_a_m      = %01_111_110
  mov_a_a      = %01_111_111



  stax_mask    = %11101111
  stax         = %00000010
  stax_b       = %00000010
  stax_d       = %00010010

  ldax_mask    = %11101111
  ldax         = %00001010
  ldax_b       = %00001010
  ldax_d       = %00011010

  'Register or memory to accumulator instructions
  rmtoacc_mask     = %11_000_000
  rmtoacc          = %10_000_000
  rmtoacc_op_mask  = %00_111_000
  rmtoacc_reg_mask = %00_000_111

  add_b            = %10_000_000
  add_c            = %10_000_001
  add_d            = %10_000_010
  add_e            = %10_000_011
  add_h            = %10_000_100
  add_l            = %10_000_101
  add_m            = %10_000_110
  add_a            = %10_000_111
         
  adc_b            = %10_001_000
  adc_c            = %10_001_001
  adc_d            = %10_001_010
  adc_e            = %10_001_011
  adc_h            = %10_001_100
  adc_l            = %10_001_101
  adc_m            = %10_001_110
  adc_a            = %10_001_111

  sub_b            = %10_010_000
  sub_c            = %10_010_001
  sub_d            = %10_010_010
  sub_e            = %10_010_011
  sub_h            = %10_010_100
  sub_l            = %10_010_101
  sub_m            = %10_010_110
  sub_a            = %10_010_111

  sbb_b            = %10_011_000
  sbb_c            = %10_011_001
  sbb_d            = %10_011_010
  sbb_e            = %10_011_011
  sbb_h            = %10_011_100
  sbb_l            = %10_011_101
  sbb_m            = %10_011_110
  sbb_a            = %10_011_111

  ana_b            = %10_100_000
  ana_c            = %10_100_001
  ana_d            = %10_100_010
  ana_e            = %10_100_011
  ana_h            = %10_100_100
  ana_l            = %10_100_101
  ana_m            = %10_100_110
  ana_a            = %10_100_111

  xra_b            = %10_101_000
  xra_c            = %10_101_001
  xra_d            = %10_101_010
  xra_e            = %10_101_011
  xra_h            = %10_101_100
  xra_l            = %10_101_101
  xra_m            = %10_101_110
  xra_a            = %10_101_111
                                
  ora_b            = %10_110_000
  ora_c            = %10_110_001
  ora_d            = %10_110_010
  ora_e            = %10_110_011
  ora_h            = %10_110_100
  ora_l            = %10_110_101
  ora_m            = %10_110_110
  ora_a            = %10_110_111

  cmp_b            = %10_111_000
  cmp_c            = %10_111_001
  cmp_d            = %10_111_010
  cmp_e            = %10_111_011
  cmp_h            = %10_111_100
  cmp_l            = %10_111_101
  cmp_m            = %10_111_110
  cmp_a            = %10_111_111

  'Rotate accumulator instructions
  rot_mask = %111_00_111
  rot      = %000_00_111
  rlc      = %000_00_111
  rrc      = %000_01_111
  ral      = %000_10_111
  rar      = %000_11_111
  op_rlc   = 0
  op_rrc   = 1
  op_ral   = 2
  op_rar   = 3

  'PUSH instuctions
  push_mask = %11_00_1111
  push      = %11_00_0101
  push_b    = %11_00_0101
  push_d    = %11_01_0101
  push_h    = %11_10_0101
  push_psw  = %11_11_0101

  'POP instructions
  pop_mask  = %11_00_1111
  pop       = %11_00_0001
  pop_b     = %11_00_0001
  pop_d     = %11_01_0001
  pop_h     = %11_10_0001
  pop_psw   = %11_11_0001

  'DAD insructions
  dad_mask = %11_00_1111
  dad      = %00_00_1001
  dad_b    = %00_00_1001
  dad_d    = %00_01_1001
  dad_h    = %00_10_1001
  dad_sp   = %00_11_1001

  'DCX instructions
  dcx_mask = %11_00_1111
  dcx      = %00_00_1011
  dcx_b    = %00_00_1011
  dcx_d    = %00_01_1011
  dcx_h    = %00_10_1011
  dcx_sp   = %00_11_1011

  'INX instructions
  inx_mask = %11_00_1111
  inx      = %00_00_0011
  inx_b    = %00_00_0011
  inx_d    = %00_01_0011
  inx_h    = %00_10_0011
  inx_sp   = %00_11_0011

  lxi_mask = %11_00_1111
  lxi      = %00_00_0001
  lxi_b    = %00_00_0001
  lxi_d    = %00_01_0001
  lxi_h    = %00_10_0001
  lxi_sp   = %00_11_0001

  
  mvi_mask = %11_000_111
  mvi      = %00_000_110
  mvi_b    = %00_000_110
  mvi_c    = %00_001_110
  mvi_d    = %00_010_110
  mvi_e    = %00_011_110
  mvi_h    = %00_100_110
  mvi_l    = %00_101_110
  mvi_m    = %00_110_110
  mvi_a    = %00_111_110

  accimm_mask = %11_000_111
  accimm      = %11_000_110
  adi         = %11_000_110
  aci         = %11_001_110
  sui         = %11_010_110
  sbi         = %11_011_110
  ani         = %11_100_110
  xri         = %11_101_110
  ori         = %11_110_110
  cpi         = %11_111_110

  diradd_mask = %111_00_111
  diradd      = %001_00_010
  dirop_sta   = %10
  dirop_lda   = %11
  dirop_shdl  = %00
  dirop_lhdl  = %01  
  sta         = %001_10_010
  lda         = %001_11_010
  shld        = %001_00_010
  lhld        = %001_01_010  

  jmp_mask    = %11_000_111
  jmp_cond    = %11_000_010
  jmp_        = %11_000_011
  jnz         = %11_000_010
  jz          = %11_001_010
  jnc         = %11_010_010
  jc          = %11_011_010
  jpo         = %11_100_010
  jpe         = %11_101_010
  jp          = %11_110_010
  jm          = %11_111_010
  
  call_mask   = %11_000_111
  call_cond   = %11_000_100
  call_       = %11_001_101
  cnz         = %11_000_100
  cz          = %11_001_100
  cnc         = %11_010_100
  cc          = %11_011_100
  cpo         = %11_100_100
  cpe         = %11_101_100
  cp          = %11_110_100
  cm          = %11_111_100

  ret_mask    = %11_000_111
  ret_cond    = %11_000_000
  ret_        = %11_001_001
  rnz         = %11_000_000
  rz          = %11_001_000
  rnc         = %11_010_000
  rc          = %11_011_000
  rpo         = %11_100_000
  rpe         = %11_101_000
  rp          = %11_110_000
  rm          = %11_111_000

  'Condition fields for conditional jumps, calls and returns
  cond_nz     = %000
  cond_z      = %001
  cond_nc     = %010
  cond_c      = %011
  cond_po     = %100
  cond_pe     = %101
  cond_p      = %110
  cond_m      = %111

  rst_mask    = %11_000_111
  rst         = %11_000_111

  io_mask     = %11110111
  ei_di_mask  = %11110111

  
  noop  = %00000000
  halt  = %01110110

  out   = %11010011
  in    = %11011011
  di    = %11110011
  ei    = %11111011

  xchg  = %11101011
  xthl  = %11100011
  sphl  = %11111001
  
  'Jump instructions
  pchl  = %11101001

  cma   = %00101111
  daa   = %00100111

  'Program status word bits
  sign_bit   = %10000000
  zero_bit   = %01000000
  aux_bit    = %00010000
  parity_bit = %00000100
  carry_bit  = %00000001


VAR

  long  cog


PUB start(cpupar) : okay

'' Start Altair CPU simulator - starts a cog
'' returns false if no cog available
''
''   cpupar = pointer to CPU parameters

  ram_base := long[cpupar]
  rom_base := long[cpupar + 4]
  ios_op   := long[cpupar + 8]
  ios_port := ios_op + 1
  ios_data := ios_op + 2

  reg_ptr  := @registers
  dispatch_tab_addr := @dispatch_table
  
  stop
  okay := cog := cognew(@run, cpupar) + 1


PUB stop

'' Stop Altair CPU simulator - frees a cog

  if cog
    cogstop(cog~ - 1)


DAT
                        org     0
run
do_nop
                        call    #instruction_fetch        'Get the next instruction into 'ins'
                   
                        mov     op, ins                   'Extract 3 bits from instruction bits 3 to 5 as op
                        shr     op, #3
                        and     op, #%00000111

                        mov     reg_pair, ins             'Extract register pair code from instruction into reg_pair
                        and     reg_pair, #rp_mask
                        shr     reg_pair, #3              'N.B. shift by 3 not 4 as reg_pair indexes the first reg in register file       

                        mov     reg_no, ins               'Extract register number from instruction into reg
                        and     reg_no, #%00_000_111

                        mov     scratch,ins               'Dispatch table look up
                        shl     scratch, #1               'mul by 2 to get a WORD offset into the table
                        add     scratch, dispatch_tab_addr
                        rdword  scratch, scratch
                        jmp     scratch                   'Jump to the instruction handler



data          long  0
address       long  0
pointer       long  0
ram_size      long  1024 * 24
rom_start     long  $FF00

'Read the next word from pc and pc + 1 and place it in "imm_word"
word_fetch              call    #instruction_fetch        'Read immediate byte into 'ins'
                        mov     imm_word, ins             'Get it to top byte of "mem_ptr"
                        ror     imm_word, #8
                        call    #instruction_fetch        'Read immediate byte into 'ins'
                        or      imm_word, ins             'Get it into low byte of "mem_ptr"
                        rol     imm_word, #8              'Rotate them into correct positions
word_fetch_ret          ret                      
imm_word long 0

'Read an instruction from byte at pc into ins, increment pc by 1. 
instruction_fetch       mov     address, pc               'Set address param for read_memory
                        call    #read_memory              'Read the byte there 
                        mov     ins, data                 'Return the instruction in ins
                        and     ins, #$FF
                        add     pc, #1                    'Increment the program counter
                        and     pc, word_mask             'Ensure it is only 16 bits still (roll round)
instruction_fetch_ret   ret


'Read a byte from location given by "address" into byte "data" ("address" is changed)
read_memory             cmp     address, ram_size wc,wz   'Is the given address within the RAM range ?
              if_b      mov     pointer, ram_base         'Point to the RAM
              if_b      add     pointer, address          'Add address offset
              if_b      rdbyte  data, pointer             'Read the byte from there
read_memory_ret if_b    ret                               'Return or fall through to read the ROM
read_rom                cmp     address, rom_start wc,wz  'Is the given address within the ROM range ?
              if_b      jmp     #read_memory_ret          'address below rom_base => return with "data" undefined
                        sub     address, rom_start        'address becomes offset into ROM from start of ROM                  
                        mov     pointer, rom_base         'Point to the ROM
                        add     pointer, address          'Add address offset                     
                        rdbyte  data, pointer             'Read the byte from there
                        cmp     0,0                       'Ensure carry clear for the conditional return
read_rom_end            jmp     #read_memory_ret          'Return


'Write the byte "data" to location given by "address"
write_memory            cmp     address, ram_size wc,wz   'Is the given address within the RAM range ?
              if_b      mov     pointer, ram_base         'Point to the RAM
              if_b      add     pointer, address          'Add address offset
              if_b      wrbyte  data, pointer             'Write the byte there
write_memory_ret        ret

do_stc                  or      psw_reg, #carry_bit
                        jmp     #run

do_cmc                  xor     psw_reg, #carry_bit
                        jmp     #run
           
do_daa                  jmp     #run


do_halt                 jmp     #do_halt

do_di
do_ei
do_ei_di                mov     int_enable_flag, op
                        jmp     #run


do_in
do_out
do_io                   call    #instruction_fetch        'Yes: Read immediate byte of instruction
                        wrbyte  ins, ios_port
                        wrbyte  a_reg, ios_data
                        wrbyte  op, ios_op
                                  
        
io_not_rdy              rdbyte  scratch, ios_op wz
              if_nz     jmp     #io_not_rdy
                        rdbyte  a_reg, ios_data
                        jmp     #run

ios_port                long    0
ios_op                  long    0
ios_data                long    0

do_cma                  xor     a_reg, #%11111111
                        jmp     #run


do_stax_b               mov     address, b_reg
                        shl     address, #8
                        add     address, c_reg
                        jmp     #do_stax
           

do_stax_d               mov     address, d_reg
                        shl     address, #8
                        add     address, e_reg
do_stax                 mov     data, a_reg
                        call    #write_memory
                        jmp     #run


do_ldax_b               mov     address, b_reg
                        shl     address, #8
                        add     address, c_reg
                        jmp     #do_ldax


do_ldax_d               mov     address, d_reg
                        shl     address, #8
                        add     address, e_reg


do_ldax                 call    #read_memory              'Read byte from "address" into "data"
                        mov     a_reg, data
                        jmp     #run


do_inr                  mov     scratch, #1               'Set up a one for increment
                        jmp     #bump


do_dcr                  mov     scratch, #%1111_1111      'Set up a minus one for decrement
bump                    mov     reg_no, op                'Extract register number from instruction op field
                        call    #get_reg_m                'Get specified reg from reg file or memory(hl) into alu
                        mov     op, #8                    'op = 8 is alu add leaving carry alone
                        call    #alu_op                   'Bump and set flags
                        call    #put_reg_m                'Put alu to specified reg in reg file or memory(hl)
                        jmp     #run


do_ral                  rol     a_reg, #1
                        test    psw_reg, #carry_bit wz
              if_nz     or      a_reg, #1
                        test    a_reg, #%1_0000_0000 wz
                        jmp     #do_ral_tail
             
do_rar                  ror     a_reg, #1 wc
                        test    psw_reg, #carry_bit wz
              if_nz     or      a_reg, #%10000000
                        jmp     #do_rar_tail

do_rlc                  rol     a_reg, #1
                        test    a_reg, #%1_0000_0000 wz
              if_nz     or      a_reg, #1
do_ral_tail
              if_nz     or      psw_reg, #carry_bit    
              if_z      andn    psw_reg, #carry_bit    
                        jmp     #run

do_rrc                  ror     a_reg, #1 wc 
              if_c      or      a_reg, #%10000000
do_rar_tail
              if_c      or      psw_reg, #carry_bit    
              if_nc     andn    psw_reg, #carry_bit
                        jmp     #run


do_mov                  cmp     ins, #halt wz             'Moving mem to mem is n ot allowed and defined as HALT
              if_z      jmp     #do_halt
                        call    #get_reg_m                'Get specified reg from reg file into alu
                        jmp     #mov_end

do_mvi                  call    #instruction_fetch        'Get the instructions imediate data  
                        mov     alu, ins            
mov_end                 mov     reg_no, op                'op contains dest register of mvi instructions
                        call    #put_reg_m
                        jmp     #run

do_rmtoacc              call    #get_reg_m                'Get specified reg from reg file into alu
                        mov     scratch, alu              'alu into scratch
                        jmp     #acc_end             


do_accimm
do_adi
do_aci
do_sui
do_sbi
do_ani
do_xri
do_ori
do_cpi                  call    #instruction_fetch        'Get the instructions imediate data
                        mov     scratch, ins              'Immediate data is alu operand in scratch

acc_end                 mov     alu, a_reg
                        call    #alu_op
                        cmp     op, #op_cmp wz
              if_nz     mov     a_reg,alu
                        jmp     #run

do_lhld                 call    #word_fetch               'Get instructions immediate word into "imm_word"
                        mov     address, imm_word         'Set "address" param of read_memory to the imediate word
                        call    #read_memory              'Read byte from "address" into "data"
                        mov     l_reg, data        
                        add     address, #1               'Address the next byte
                        call    #read_memory              'Read byte from "address" into "data"
                        mov     h_reg, data
                        jmp     #run

do_shld                 call    #word_fetch               'Get instructions immediate word into "imm_word"
                        mov     address, imm_word         'Set "address" param of read_memory to the imediate word
                        mov     data, l_reg               'Set "data" param of write_memory to "l" reg
                        call    #write_memory             'Write out the "l" reg to "address"
                        add     address, #1            
                        mov     data, h_reg               'Set "data" param of write_memory to "h" reg
                        call    #write_memory             'Write out the "h" reg to "address"
                        jmp     #run                 

do_sta                  call    #word_fetch               'Get instructions immediate word into "imm_word"
                        mov     address, imm_word         'Set "address" param of write_memory to "mem_ptr"
                        mov     data, a_reg               'Set "data" param of write_memory to "a" reg
                        call    #write_memory             'Write out the "a" reg to "address"
                        jmp     #run

do_lda                  call    #word_fetch               'Get instructions immediate word into "imm_word"
                        mov     address, imm_word         'Set "address" param of write_memory to the immediate word
                        call    #read_memory              'Read a byte from "address" into "data"
                        mov     a_reg, data               'Get the read byte in "a" reg
                        jmp     #run

             
do_push                 call    #get_reg_pair             'Get specified register pair into 'scratch'
                        call    #push_word                'Push the register pair
                        jmp     #run

do_pop                  call    #pop_word                 'Pop word from stack into 'scratch'
                        call    #put_reg_pair             'Put 'scratch' to specified register pair_pair
                        jmp     #run
                                     
do_dad                  call    #get_reg_pair_sp          'Get specified register pair into 'scratch'
                        mov     add_hl_t, h_reg
                        shl     add_hl_t, #8
                        add     add_hl_t, l_reg
                        add     add_hl_t, scratch
                        test    add_hl_t, bit_16 wz
              if_nz     or      psw_reg, #carry_bit
                        mov     l_reg, add_hl_t
                        and     l_reg, #%11111111
                        shr     add_hl_t, #8
                        mov     h_reg, add_hl_t
                        and     h_reg, #%11111111
                        jmp     #run
                        
add_hl_t                long     0
                        

do_inx                  call    #get_reg_pair_sp          'Get specified register pair into 'scratch'
                        add     scratch, #1
                        jmp     #lxi_end

             
do_dcx                  call    #get_reg_pair_sp          'Get specified register pair into 'scratch'
                        sub     scratch, #1
                        jmp     #lxi_end


do_lxi                  call    #instruction_fetch        'Read immediate byte into 'ins'
                        mov     scratch, ins              'Get it to top byte of scratch
                        ror     scratch, #8
                        call    #instruction_fetch        'Read immediate byte into 'ins'
                        or      scratch, ins              'Get it into low byte of scratch
                        rol     scratch, #8               'Rotate them into correct positions
                        
lxi_end                 call    #put_reg_pair_sp          'Put "scratch" to specified register pair
                        jmp     #run

do_xchg                 xor     h_reg, d_reg              'Exchage HL with DE
                        xor     d_reg, h_reg
                        xor     h_reg, d_reg
                        xor     l_reg, e_reg
                        xor     e_reg, l_reg
                        xor     l_reg, e_reg
                        jmp     #run

do_xthl                 mov     address, sp               'Set address param for read_memory to stack pointer
                        call    #read_memory              'Read byte from "address" into "data"
                        xor     data, l_reg               'Swap l reg and data
                        xor     l_reg, data
                        xor     data, l_reg
                        call    #write_memory             'Write the byte "data" to location "address" 
                        add     address, #1               'Set address param for read_memory to next stack location up
                        call    #read_memory              'Read byte from "address" into "data"
                        xor     data, h_reg               'Swap h reg and data
                        xor     h_reg, data
                        xor     data, h_reg
                        call    #write_memory             'Write the byte "data" to location "address" 
                        jmp     #run

do_sphl                 mov     sp, h_reg
                        shl     sp, #8
                        add     sp, l_reg
                        jmp     #run

do_pchl                 mov     pc, h_reg
                        shl     pc, #8
                        add     pc, l_reg
                        jmp     #run

'Conditional return instructions
do_rnz
do_rz
do_rnc
do_rc
do_rpo
do_rpe
do_rp
do_rm
do_ret_cond             movs    fytch, #do_ret
                        jmp     #do_conditional

'Conditional call instructions
do_cnz
do_cz
do_cnc
do_cc
do_cpo
do_cpe
do_cp
do_cm
do_call_cond            call    #word_fetch               'Get instructions immediate word into "imm_word"
                        movs    fytch, #do_call_end
                        jmp     #do_conditional

'Conditional jmp instructions
do_jnz
do_jz
do_jnc
do_jc
do_jpo
do_jpe
do_jp
do_jm
do_jmp_cond             call    #word_fetch               'Get instructions immediate word into "imm_word"
                        movs    fytch, #do_jmp_end
             
do_conditional
                        mov     scratch, psw_reg          'Get curent state of the flags 
                        test    op, #1 wz                 'wc here keeps low bit of condition indicating inverse test
              if_z      xor     scratch, #%11111111       'Invert flags for not zero, no carry, parity even and plus conditions            

                        andn    op, #1                    'Remove low bit
                        shl     op, #1

                        movs    jmp_test, #test_cond
                        add     jmp_test, op
                        nop 
jmp_test                jmp     #test_cond
       
test_cond               test    scratch, #zero_bit wz
              if_nz     jmp     fytch
                        jmp     #run
                        nop                               'This pads out steps between tests to 4 longs so the jumps work !!!

                        test    scratch, #carry_bit wz
              if_nz     jmp     fytch
                        jmp     #run
                        nop                               'This pads out steps between tests to 4 longs so the jumps work !!!

                        test    scratch, #parity_bit wz
              if_nz     jmp     fytch
                        jmp     #run
                        nop                               'This pads out steps between tests to 4 longs so the jumps work !!!

                        test    scratch, #sign_bit wz
              if_z      jmp     #run

fytch         if_nz     jmp     #do_jmp             
                        jmp     #run                

do_jmp                  call    #word_fetch               'Get instructions immediate word into "imm_word"
do_jmp_end              mov     pc, imm_word              'Set program counter to the immediate address
                        jmp     #run

do_call                 call    #word_fetch               'Get instructions immediate word into "imm_word"
do_call_end             mov     scratch, pc               'Push this return address to stack
                        call    #push_word
                        jmp     #do_jmp_end          

do_ret                  call    #pop_word                 'Pop the return address from the stack into 'scratch'
                        mov     pc, scratch               'Set program counter to popped address
                        jmp     #run
                  
do_rst                  mov     pc, #0                    'RST, Restart at PC = 0 + op field shifted up by 3
                        shl     op, #3
                        add     pc, op
                        jmp     #run

'All undocumented 8085 opcodes come here
do_undocumented         jmp     #run
                        

'Pop a word from the stack into 'scratch'
pop_word                mov     address, sp               'Set "address" param or read_word to sp    
                        call    #read_memory              'Read a byte from the stack into "data"
                        mov     scratch, data             'Put it to top byte of scratch
                        ror     scratch, #8               
                        add     sp, #1                    'Bump up the stack pointer
                        and     sp, word_mask             'Ensure still 16 bits (roll over)
                        mov     address, sp               'Set "address" param or read_word to sp
                        call    #read_memory              'Read a byte from the stack into "data"
                        or      scratch, data             'Put it to low byte of scratch
                        rol     scratch, #8               'Roll bytes into correct order
                        add     sp, #1                    'Bump up the stack pointer
                        and     sp, word_mask             'Ensure still 16 bits (roll over)                        
pop_word_ret            ret


'Push a word from 'scratch' onto the stack
push_word               sub     sp, #1                    'Decrement stack pointer first
                        and     sp, word_mask             'Ensure still 16 bit
                        mov     address,sp                'Set "address" parameter of write_memory to stack pointer
                        mov     data,scratch              'Set "data" parameter of write_memory to "scratch"
                        ror     data, #8                  'Get high byte into low end
                        call    #write_memory             'Write the high byte to the stack
                        sub     sp, #1                    'Decrement stack pointer again
                        and     sp, word_mask             'Ensure still 16 bit
                        mov     address,sp                'Set "address" parameter of write_memory to stack pointer
                        mov     data, scratch             'Set "data" parameter of write_memory to "scratch" 
                        call    #write_memory             'Write the low byte to the stack
push_word_ret           ret

             
'Perform the alu operation specified by 'op'
alu_op                  and     alu, #255                 'Ensure both operands are in range
                        and     scratch, #255

                        movs    fotch, #alu_op_tbl             
                        add     fotch, op                 'Make a jump to the required alu op handler
alu                     nop                               'Essential nop to cater for propeller pipeline
                                                          'It also holds the 'alu' variable !!!
fotch                   jmp     #alu_op_tbl            
alu_op_tbl              jmp     #alu_add
                        jmp     #alu_adc
                        jmp     #alu_sub
                        jmp     #alu_sbb
                        jmp     #alu_and
                        jmp     #alu_xor
                        jmp     #alu_or
                        jmp     #alu_sub                  'Compares (CMP & CPI) are subtracts but the result is discarded
                        'jmp     #alu_add_ncj
alu_add_ncj
                        call    #alu_add_nc
                        jmp     #alu_op_tail

                        
'Add 'scratch' and carry to 'alu' and set carry flag  accordingly
alu_adc                 test    psw_reg, #carry_bit wz
              if_nz     add     scratch, #1

      
'Add 'scratch' to 'alu' and set carry flag and aux carry accordingly
alu_add
                        call    #alu_add_nc
              if_z      test    alu, #256 wz
              if_nz     or      psw_reg, #carry_bit
              if_z      andn    psw_reg, #carry_bit
                        jmp     #alu_op_tail

little_alu              long    0

                           
alu_add_nc              mov     little_alu, alu
                        add     alu, scratch
                        and     little_alu, #%1111
                        add     little_alu, scratch
                        test    little_alu, #%10000 wz
              if_nz     or      psw_reg, #aux_bit
              if_z      andn    psw_reg, #aux_bit      
alu_add_nc_ret          ret


'Subtract 'scratch' and carry from 'alu' and set carry flag accordingly
alu_sbb                 test    psw_reg, #carry_bit wz
              if_nz     add     scratch, #1

'Subtract 'scratch' from 'alu' and set carry flag  accordingly
alu_sub                 neg     scratch, scratch
                        jmp     #alu_add
             
'Bitwise AND of 'scratch' and 'alu' and clear carry flag
alu_and                 and     alu, scratch
                        jmp     #clear_carry

'Bitwise ExOR of 'scratch' and 'alu' and clear carry flag 
alu_xor                 xor     alu, scratch
                        jmp     #clear_carry

'Bitwise OR of 'scratch' and 'alu' and clear carry flag 
alu_or                  or      alu, scratch
clear_carry             andn    psw_reg, #carry_bit       'Clear the carry flag

alu_op_tail             andn    psw_reg, #%11000100       'Clear zero, sign and parity bits
                        and     alu, #255 wz, wc
              if_z      or      psw_reg, #zero_bit
              if_nc     or      psw_reg, #parity_bit
                        test    alu, #128 wz
              if_nz     or      psw_reg, #sign_bit
alu_op_ret              ret

'Put the 'alu' to the register 'reg' in the register file or memory (hl)
put_reg_m               cmp     reg_no, #reg_m wz         'Is destination a register or memory (hl)
              if_nz     jmp     #put_reg
                        mov     data,alu                  'Set "data" parameter for write_memory to "alu"
                        mov     address, h_reg            'Make an address out of h and l
                        shl     address, #8
                        add     address, l_reg
                        call    #write_memory             'Write "alu" out to "address"
                        jmp     #put_reg_m_ret     

'Put the 'alu' to the register 'reg' in the register file
put_reg                 movd    fitch, #registers         'Use reg to index into registers
                        shl     reg_no, #9
                        add     fitch, reg_no
                        and     alu, #255                 'Ensure register will be in range
fitch                   mov     registers, alu            'Get specified registers contents 
put_reg_m_ret
put_reg_ret             ret

put_reg_pair_sp         cmp     reg_pair, #%110 wz
              
              if_z      mov     sp, scratch
              if_z      jmp     #put_reg_pair_sp_ret

'Put 'scratch' into register pair specified by 'reg_pair'
put_reg_pair            cmp     reg_pair, #%110 wz        'If reg_pair is 'psw' adjust it to index A and PWS in reg file
              if_z      add     reg_pair, #1 
                        mov     reg_tmp, #registers       'Point to the register file base
                        add     reg_tmp, reg_pair         'Offset to required register pair
                        movd    put_r_high, reg_tmp       'Modify our put low byte inst. to write that reg.
                        add     reg_tmp, #1               'Point to next reg
                        movd    put_r_low, reg_tmp        'Modify our put high byte inst. to write that reg.
                        mov     reg_tmp, scratch
                        and     reg_tmp, #255
put_r_low               mov     registers, reg_tmp        'Put the low byte to it's reg
                        shr     scratch, #8               'Get the high byte
                        and     scratch, #255
put_r_high              mov     registers, scratch        'and put it to it's reg
put_reg_pair_sp_ret
put_reg_pair_ret        ret
                        

reg_tmp                 long    0

'Get the register 'reg' from the register file or memory (hl) into 'alu' 
get_reg_m               cmp     reg_no, #reg_m wz         'Is this a register or memory (hl) access ?
              if_nz     jmp     #get_reg
                        mov     address, h_reg            'Make an address out of h and l
                        shl     address, #8
                        add     address, l_reg
                        call    #read_memory              'Read byte from "address" into "data"
                        mov     alu, data
                        jmp     #get_reg_m_ret


'Get the register 'reg' from the register file into 'alu' 
get_reg                 movs    fetch, #registers         'Use reg to index into registers
                        add     fetch, reg_no
                        nop                               'Essential nop to cater for propeller pipeline
fetch                   mov     alu, registers            'Get specified registers contents 
get_reg_m_ret
get_reg_ret             ret


get_reg_pair_sp         cmp     reg_pair, #%110 wz
              if_z      mov     scratch, sp
              if_z      jmp     #get_reg_pair_sp_ret

'Get the register pair specified by 'reg_pair' into 'scratch'
get_reg_pair
                        cmp     reg_pair, #%110 wz        'If reg_pair is 'psw' adjust it to index A and PWS in reg file
              if_z      add     reg_pair, #1 
                        mov     reg_tmp, #registers       'Point to the register file base
                        add     reg_tmp, reg_pair         'Offset to required register pair
                        movs    :get_r_high, reg_tmp      'Modify our get low byte inst. to read that reg.
                        add     reg_tmp, #1               'Point to next reg
                        movs    :get_r_low, reg_tmp       'Modify our get high byte inst. to write that reg.
:get_r_high             mov     scratch, registers        'Get the high reg and 
                        shl     scratch, #8               'move to the high byte
:get_r_low              add     scratch, registers        'Get the reg into low byte
get_reg_pair_sp_ret
get_reg_pair_ret        ret



' Display a byte on the leds for one second
{{
display      mov             dira, pins                   'Set Pins to output
             shl             disp, #16                    'Get data into right bits
             mov             outa, disp                   'Display it
             mov             time, cnt                    'Calculate delay time
             add             time, delay                  'Set 1 second delay here
             waitcnt         time, 0                      'Wait
display_ret  ret
pins    long %00000000_11111111_00000000_00000000
time    long 0
delay   long 80_000_000 * 5
disp    long 0
}}

reg_pair      long    0


bit_16        long    %00000000_00000001_00000000_00000000
bit_31        long    %10000000_00000000_00000000_00000000
word_mask     long    %00000000_00000000_11111111_11111111


'8080 register file, DO NOT change the order of these or add anything in between
registers
b_reg         long    0
c_reg         long    0
d_reg         long    0
e_reg         long    0
h_reg         long    0
l_reg         long    0
m_reg         long    0                                   'Not used just holds a space so that get_reg works
a_reg         long    0
psw_reg       long    %00000000


'Other 8080 regs
sp            long    0
pc            long    0

'Interrupt enable flip flop                               'Start in the boot rom
int_enable_flag         long    0

{{
dump_regs               mov     scratch, #12              
dmp                     wrlong  registers,reg_ptr 'modify dst to get next register 
                        add     reg_ptr,#4
                        add     dmp,dst_1
                        sub     scratch,#1
                        tjnz    scratch,#dmp                             
dump_regs_ret           ret

dst_1         long %1000000000
}}
reg_ptr       long 0
   
'Misc stuff
ins           long    0
ram_base      long    0
rom_base      long    0
scratch       long    0
mem_ptr       long    0
op            long    0
reg_no        long    0

dispatch_tab_addr    long    0


'61 longs free !!!  + There are 5 nops to use up yet !
long 0
long 0
long 0
long 0
long 0
long 0
long 0
long 0
long 0
long 0

long 0
long 0
long 0
long 0
long 0
long 0
long 0
long 0
long 0
long 0






                        fit

                        
'Instruction dispatch table.
'Notice that this is used from HUB RAM as it will not fit in the cog. There is a small speed penalty
'in doing this (rdword) but not much compared to total instruction execution time.
dispatch_table
{00000000}     word     do_nop  
{00000001}     word     do_lxi  
{00000010}     word     do_stax_b  
{00000011}     word     do_inx
{00000100}     word     do_inr
{00000101}     word     do_dcr  
{00000110}     word     do_mvi  
{00000111}     word     do_rlc  
{00001000}     word     do_undocumented  
{00001001}     word     do_dad  
{00001010}     word     do_ldax_b  
{00001011}     word     do_dcx  
{00001100}     word     do_inr  
{00001101}     word     do_dcr  
{00001110}     word     do_mvi  
{00001111}     word     do_rrc  
{00010000}     word     do_undocumented  
{00010001}     word     do_lxi  
{00010010}     word     do_stax_d  
{00010011}     word     do_inx  
{00010100}     word     do_inr  
{00010101}     word     do_dcr  
{00010110}     word     do_mvi  
{00010111}     word     do_ral  
{00011000}     word     do_undocumented  
{00011001}     word     do_dad  
{00011010}     word     do_ldax_d  
{00011011}     word     do_dcx  
{00011100}     word     do_inr  
{00011101}     word     do_dcr  
{00011110}     word     do_mvi  
{00011111}     word     do_rar  
{00100000}     word     do_undocumented  
{00100001}     word     do_lxi  
{00100010}     word     do_shld  
{00100011}     word     do_inx  
{00100100}     word     do_inr  
{00100101}     word     do_dcr  
{00100110}     word     do_mvi  
{00100111}     word     do_daa  
{00101000}     word     do_undocumented  
{00101001}     word     do_dad  
{00101010}     word     do_lhld  
{00101011}     word     do_dcx  
{00101100}     word     do_inr  
{00101101}     word     do_dcr  
{00101110}     word     do_mvi  
{00101111}     word     do_cma  
{00110000}     word     do_undocumented  
{00110001}     word     do_lxi  
{00110010}     word     do_sta  
{00110011}     word     do_inx  
{00110100}     word     do_inr  
{00110101}     word     do_dcr  
{00110110}     word     do_mvi  
{00110111}     word     do_stc  
{00111000}     word     do_undocumented  
{00111001}     word     do_dad  
{00111010}     word     do_lda  
{00111011}     word     do_dcx  
{00111100}     word     do_inr  
{00111101}     word     do_dcr  
{00111110}     word     do_mvi  
{00111111}     word     do_cmc  

'All the mov instructions
{01000000}     word     do_mov  
{01000001}     word     do_mov  
{01000010}     word     do_mov  
{01000011}     word     do_mov
{01000100}     word     do_mov
{01000101}     word     do_mov  
{01000110}     word     do_mov  
{01000111}     word     do_mov
{01001000}     word     do_mov  
{01001001}     word     do_mov  
{01001010}     word     do_mov 
{01001011}     word     do_mov  
{01001100}     word     do_mov  
{01001101}     word     do_mov  
{01001110}     word     do_mov  
{01001111}     word     do_mov
{01010000}     word     do_mov
{01010001}     word     do_mov  
{01010010}     word     do_mov  
{01010011}     word     do_mov  
{01010100}     word     do_mov  
{01010101}     word     do_mov  
{01010110}     word     do_mov  
{01010111}     word     do_mov
{01011000}     word     do_mov  
{01011001}     word     do_mov  
{01011010}     word     do_mov  
{01011011}     word     do_mov  
{01011100}     word     do_mov  
{01011101}     word     do_mov  
{01011110}     word     do_mov  
{01011111}     word     do_mov
{01100000}     word     do_mov  
{01100001}     word     do_mov  
{01100010}     word     do_mov  
{01100011}     word     do_mov  
{01100100}     word     do_mov  
{01100101}     word     do_mov  
{01100110}     word     do_mov  
{01100111}     word     do_mov
{01101000}     word     do_mov  
{01101001}     word     do_mov  
{01101010}     word     do_mov  
{01101011}     word     do_mov  
{01101100}     word     do_mov  
{01101101}     word     do_mov  
{01101110}     word     do_mov  
{01101111}     word     do_mov
{01110000}     word     do_mov  
{01110001}     word     do_mov  
{01110010}     word     do_mov  
{01110011}     word     do_mov  
{01110100}     word     do_mov  
{01110101}     word     do_mov  
{01110110}     word     do_mov  
{01110111}     word     do_mov
{01111000}     word     do_mov  
{01111001}     word     do_mov  
{01111010}     word     do_mov  
{01111011}     word     do_mov  
{01111100}     word     do_mov  
{01111101}     word     do_mov  
{01111110}     word     do_mov  
{01111111}     word     do_mov  

'All the reg/mem to accumulator instructions
{10000000}     word     do_rmtoacc  
{10000001}     word     do_rmtoacc
{10000010}     word     do_rmtoacc
{10000011}     word     do_rmtoacc
{10000100}     word     do_rmtoacc
{10000101}     word     do_rmtoacc
{10000110}     word     do_rmtoacc
{10000111}     word     do_rmtoacc
{10001000}     word     do_rmtoacc
{10001001}     word     do_rmtoacc
{10001010}     word     do_rmtoacc
{10001011}     word     do_rmtoacc
{10001100}     word     do_rmtoacc
{10001101}     word     do_rmtoacc
{10001110}     word     do_rmtoacc
{10001111}     word     do_rmtoacc
{10010000}     word     do_rmtoacc
{10010001}     word     do_rmtoacc
{10010010}     word     do_rmtoacc
{10010011}     word     do_rmtoacc
{10010100}     word     do_rmtoacc
{10010101}     word     do_rmtoacc
{10010110}     word     do_rmtoacc
{10010111}     word     do_rmtoacc
{10011000}     word     do_rmtoacc
{10011001}     word     do_rmtoacc
{10011010}     word     do_rmtoacc
{10011011}     word     do_rmtoacc
{10011100}     word     do_rmtoacc
{10011101}     word     do_rmtoacc
{10011110}     word     do_rmtoacc
{10011111}     word     do_rmtoacc
{10100000}     word     do_rmtoacc
{10100001}     word     do_rmtoacc
{10100010}     word     do_rmtoacc
{10100011}     word     do_rmtoacc
{10100100}     word     do_rmtoacc
{10100101}     word     do_rmtoacc
{10100110}     word     do_rmtoacc
{10100111}     word     do_rmtoacc
{10101000}     word     do_rmtoacc
{10101001}     word     do_rmtoacc
{10101010}     word     do_rmtoacc
{10101011}     word     do_rmtoacc
{10101100}     word     do_rmtoacc
{10101101}     word     do_rmtoacc
{10101110}     word     do_rmtoacc
{10101111}     word     do_rmtoacc
{10110000}     word     do_rmtoacc
{10110001}     word     do_rmtoacc
{10110010}     word     do_rmtoacc
{10110011}     word     do_rmtoacc
{10110100}     word     do_rmtoacc
{10110101}     word     do_rmtoacc
{10110110}     word     do_rmtoacc
{10110111}     word     do_rmtoacc
{10111000}     word     do_rmtoacc
{10111001}     word     do_rmtoacc
{10111010}     word     do_rmtoacc
{10111011}     word     do_rmtoacc
{10111100}     word     do_rmtoacc
{10111101}     word     do_rmtoacc
{10111110}     word     do_rmtoacc
{10111111}     word     do_rmtoacc

{11000000}     word     do_rnz  
{11000001}     word     do_pop  
{11000010}     word     do_jnz  
{11000011}     word     do_jmp 
{11000100}     word     do_cnz 
{11000101}     word     do_push  
{11000110}     word     do_adi  
{11000111}     word     do_rst  
{11001000}     word     do_rz  
{11001001}     word     do_ret  
{11001010}     word     do_jz 
{11001011}     word     do_undocumented  
{11001100}     word     do_cz  
{11001101}     word     do_call   
{11001110}     word     do_aci  
{11001111}     word     do_rst  
{11010000}     word     do_rnc  
{11010001}     word     do_pop  
{11010010}     word     do_jnc  
{11010011}     word     do_out  
{11010100}     word     do_cnc  
{11010101}     word     do_push  
{11010110}     word     do_sui  
{11010111}     word     do_rst  
{11011000}     word     do_rc  
{11011001}     word     do_undocumented  
{11011010}     word     do_jc  
{11011011}     word     do_in  
{11011100}     word     do_cc  
{11011101}     word     do_undocumented  
{11011110}     word     do_sbi  
{11011111}     word     do_rst  
{11100000}     word     do_rpo  
{11100001}     word     do_pop  
{11100010}     word     do_jpo  
{11100011}     word     do_xthl  
{11100100}     word     do_cpo  
{11100101}     word     do_push  
{11100110}     word     do_ani  
{11100111}     word     do_rst  
{11101000}     word     do_rpe  
{11101001}     word     do_pchl  
{11101010}     word     do_jpe  
{11101011}     word     do_xchg  
{11101100}     word     do_cpe 
{11101101}     word     do_undocumented  
{11101110}     word     do_xri  
{11101111}     word     do_rst  
{11110000}     word     do_rp  
{11110001}     word     do_pop  
{11110010}     word     do_jp 
{11110011}     word     do_di  
{11110100}     word     do_cp  
{11110101}     word     do_push  
{11110110}     word     do_ori  
{11110111}     word     do_rst  
{11111000}     word     do_rm  
{11111001}     word     do_sphl  
{11111010}     word     do_jm  
{11111011}     word     do_ei  
{11111100}     word     do_cm  
{11111101}     word     do_undocumented  
{11111110}     word     do_cpi  
{11111111}     word     do_rst  

          